/************************************************************************
*									*
*	Device Driver for INMOS-B004/B008 compatible link interfaces	*
*									*
*	Copyright (C) 1993,94,95 by Christoph Niemann			*
*				 niemann@swt.ruhr-uni-bochum.de		*
*	(loosely) based on /linux/kernel/drivers/char/lp.c		*
*									*
*	Version 0.9, for Linux 1.1.83 and above, February 1995.		*
*									*
*									*
*	Module support added by Mark Bowyer <mark@lindhard.demon.co.uk> *
*									*
*	The driver may be copied under the same conditions as the	*
*	Linux operating system						*
*									*
*************************************************************************/

/* Copyright (C) 1992 by Jim Weigand, Linus Torvalds, and Michael K. Johnson */


/*
 * Do not define this symbol if you want to use the C-version of
 * the read/write functions (much slower)
 */
#undef LINK_USE_ASM

/*
 * Define this symbol if you wish to enable dma/irq support for
 * B008 compatible boards
 */
#undef B008_SUPPORT

#ifdef MODULE
#include <linux/module.h>
#include <linux/version.h>
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif

#include <linux/config.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/stat.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/linkage.h>
#include <linux/string.h>

#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/segment.h>
#ifdef B008_SUPPORT
#include <asm/irq.h>
#include <asm/dma.h>
#include <linux/mm.h>
#include <linux/signal.h>
#include <linux/interrupt.h>
#endif /* B008_SUPPORT */

#ifdef MODULE
#include "link.h"
#else
#include <linux/link.h>
#endif /* MODULE */

#include <linux/link.h>
#include <linux/dtc2.h>

/*
 * Define these symbols if you want to debug the code.
 */
#undef LINK_DEBUG 
#undef LINK_DEBUG_MORE

/*
 * If your link interface does not work with the normal i/o-commands use
 * the slower ones. But I think this should be no problem for any board
 * that uses INMOS's C012 chip.
 * Defining the symbol LINK_REALLY_SLOW_C012 will decrease the data-
 * transfer-rate from about 400 KBytes/sec to 200 KBytes/sec
 * (on my noname 486/33/16, assembler version).
 */
#undef LINK_REALLY_SLOW_C012

#ifdef LINK_REALLY_SLOW_C012
#define out		outb_p
#define in		inb_p
#define LINK_SLOW_DOWN	"outb %%al,$0x80\n\t"
#else
#define out		outb
#define in		inb
#define LINK_SLOW_DOWN	""
#endif

#ifdef LINK_DEBUG
#define DEB(x)		x
#define PRINTK(f, a...)	printk(f, ##a)
#define NO_DEB(x)
#else
#define DEB(x)
#define PRINTK(f, a...)
#define NO_DEB(x)	x
#endif

#ifdef LINK_DEBUG_MORE
#define PRINTKM(f, a...)	printk(f, ##a)
#else
#define PRINTKM(f, a...)
#endif

static int link_devices = 0;

static int link_base_addresses[LINK_NO] = { 0x150 };

static int link_boot_stat[LINK_NO] = { LINK_STATUS_RESET };

/*
 * Struct to describe the default parameters of each link interface
 */
static struct link_struct link_table[] =
	{
		 { LINK_READ_ABORT | LINK_WRITE_ABORT, 0, 0, 0, 0,
		 LINK_INIT_READ_TIMEOUT, LINK_INIT_WRITE_TIMEOUT, LINK_B004
		 },
		 { LINK_READ_ABORT | LINK_WRITE_ABORT, 0, 0, 0, 0,
		 LINK_INIT_READ_TIMEOUT, LINK_INIT_WRITE_TIMEOUT, LINK_B004
		 },
		 { LINK_READ_ABORT | LINK_WRITE_ABORT, 0, 0, 0, 0,
		 LINK_INIT_READ_TIMEOUT, LINK_INIT_WRITE_TIMEOUT, LINK_B004
		 },
	};


#ifdef LINK_USE_ASM
void link_call_sched (void)	/* doubling needed for versioned kernels
				   because of assembly call */
{
	schedule();
}
#else
#define	link_call_sched()	schedule()
#endif

static void link_call_sched_if (void)
{
	if (need_resched) schedule();
}

/****************************************************************************
 *
 * link_name()
 *
 * Return the name of the link device with minor number `minor'. This
 * function is called by the /proc filesystem
 */
static const char *link_name(dev_t minor)
{
	static char name[] = LINK_NAME"?";

	if (minor < link_devices)
	{
		name[4] = minor + '0';
		return name;
	}
	else
		return (const char*)0;
} /* link_name() */


/****************************************************************************
 *
 * static void link_reset():
 *
 * Reset transputer network.
 *
 */
static void link_reset(const int minor)
{
	PRINTK("LINK(%d) resetting transputer.\n", minor);

	out(LINK_DEASSERT_ANALYSE, LINK_ANALYSE(minor));
	udelay(LINK_DELAY);

	out(LINK_DEASSERT_RESET, LINK_RESET(minor));
	udelay(LINK_DELAY);

	out(LINK_ASSERT_RESET, LINK_RESET(minor));
	udelay(LINK_DELAY);

	out(LINK_DEASSERT_RESET, LINK_RESET(minor));
	udelay(LINK_DELAY);

	link_boot_stat[minor] = LINK_STATUS_RESET;

} /* link_reset() */

/****************************************************************************
 *
 * static void link_analyse():
 *
 * switch transputer network to analyse mode.
 *
 */
static void link_analyse( const int minor )
{
	PRINTK("LINK(%d) switching transputer to analyse mode.\n", minor);

	out( LINK_DEASSERT_ANALYSE, LINK_ANALYSE(minor));
	udelay(LINK_DELAY);

	out( LINK_ASSERT_ANALYSE, LINK_ANALYSE(minor));
	udelay(LINK_DELAY);

	out( LINK_ASSERT_RESET, LINK_RESET(minor));
	udelay(LINK_DELAY);

	out( LINK_DEASSERT_RESET, LINK_RESET(minor));
	udelay(LINK_DELAY);

	out( LINK_DEASSERT_ANALYSE, LINK_ANALYSE(minor));
	udelay(LINK_DELAY);

	link_boot_stat[minor] = LINK_STATUS_RESET | LINK_STATUS_ANALYSE;

} /* link_analyse() */


/****************************************************************************
 *
 */
 
static ssize_t read_lmem(struct file *file, char *buf, size_t count,
		loff_t *ppos)
{
	unsigned char p = *ppos & 0xff;
	int c1, c2, read;

	if (count < 0) return -EINVAL;
	if (count > 256) count = 256;
	c1 = min (count, (256 - p));
	c2 = count + p;
	read = 0;
	if (c1) memcpy_fromio(buf, (void *) (QR_PAGE2 + p), c1);
	if (c2 > 256) memcpy_fromio(buf+c1, (void *) QR_PAGE2, (c2 & 0xff));
	read += count;
	*ppos = (*ppos + read) & 0xff;
	return read;
}


static ssize_t write_lmem(struct file *file, const char *buf, size_t count,
		loff_t *ppos)
{
	unsigned char p = *ppos & 0xff;
	int c1, c2, written;

	if (count < 0) return -EINVAL;
	if (count > 256) count = 256;
	c1 = min (count, (256 - p));
	c2 = count + p;
	written = 0;
	if (c1) memcpy_toio((void *) (QR_PAGE2 + p), buf, c1);
	if (c2 > 256) memcpy_toio((void *) QR_PAGE2, buf+c1, (c2 & 0xff));
	written += count;
	*ppos = (*ppos + written) & 0xff;
	return count;
}


/****************************************************************************
 *
 */

static int link_peek (int minor, u_long addr, u_long *data)
{
	unsigned char s[8];
	int i;

	if (link_boot_stat[minor] == 0) return 1;

	s[0]=1;
	*((u_long *)(s+1)) = addr;

	for (i=0; i<5; i++) {
		while (!in(LINK_OSR(minor)) & LINK_WRITEBYTE)
			link_call_sched_if();
		out(s[i], LINK_ODR(minor));
	}
	for (i=0; i<4; i++) {
		while (!in(LINK_ISR(minor)) & LINK_READBYTE)
			link_call_sched_if();
		((char *)data)[i] = in(LINK_IDR(minor));
	}
	return 0;
}
 

static int link_poke (int minor, u_long addr, u_long data)
{
	unsigned char s[12];
	int i;

	if (link_boot_stat[minor] == 0) return 1;

	s[0]=0;
	*((u_long *)(s+1)) = addr;
	*((u_long *)(s+1)) = data;

	for (i=0; i<9; i++) {
		while (!in(LINK_OSR(minor)) & LINK_WRITEBYTE)
			link_call_sched_if();
		out(s[i], LINK_ODR(minor));
	}
	return 0;
}
 

/****************************************************************************
 *
 * static int link_poll_read() - read bytes from the link interface.
 *
 * At first, the driver checks if the link-interface is ready to send a byte
 * to the PC. If not, this check is repeated up to LINK_MAXTRY times.
 * If the link-interface is not ready after this loop, the driver sleeps for
 * LINK_TIME jiffies and then checks the link-interface again.
 * If the interface is still not ready, the driver sleeps for LINK_READ_TIMEOUT
 * jiffies. Then it checks the interface again. If it is not ready and
 * the flag LINK_READ_ABORT is set, this function returns with the number
 * of bytes read or with an error if no byte was read.
 * If the LINK_READ_ABORT flag is not set, the driver sleeps again for
 * LINK_READ_ABORT jiffies and so on.
 *
 * By default, LINK_READ_ABORT is now set.
 *
 */
static ssize_t link_read(struct file *filp, char *buf, size_t count,
		loff_t *ppos)
{
	const unsigned int minor = MINOR(filp->f_dentry->d_inode->i_rdev);
	unsigned int read_sleep = LINK_START_SLEEP;
	unsigned int read_wait  = 0;
	unsigned int read_inc = 0;
	DEB( unsigned int link_total_bytes_read = 0; )
#ifndef LINK_USE_ASM
	char 	*temp = buf;
	int	l_count;
	char	buffer[LINK_MAX_BYTES];
	int	end;
	int	i;
#else
	register char 	*temp __asm__("di") = buf;
#endif /* LINK_USE_ASM */

	PRINTKM("LINK(%d): reading %d bytes (polling mode).\n", minor, count);

	if (link_boot_stat[minor] == LINK_STATUS_RESET) {
	  u_long k;
	  if (link_peek (minor, *ppos, &k)) return -EINVAL;
	  copy_from_user(&k, buf, 4);
	  *ppos += 4;
	  return 4;
	}

	if ( count < 0)
	{
		PRINTK("LINK: invalid argument for reading: count = %d.\n", count);
		return -EINVAL;
	}

#ifdef LINK_USE_ASM
	__asm__("cld": :);

	while ( 1 )
	{
		/*
		 * This is very ugly! For a description of this algorithm
		 * see the C-version below.
		 */
		__asm__ __volatile__ (
			"movl	%4,%0\n\t"	/* we have to transfer %0 bytes */
			"push	%%es\n\t"
			"movw	%%fs,%%ax\n\t"
			"movw	%%ax,%%es\n"
		"9:	movl	%0,%4\n\t"
			"movb	$2,%%ah\n\t"
			"cmpl	%7,%4\n\t"
			"jc	6f\n\t"
			"movl	%7,%4\n"
		"6:	movl	%4,%%esi\n"
		"5:	inb	%w2,%%al\n\t"	/* ready for reading? */
			LINK_SLOW_DOWN
			"andb	%6,%%al\n\t"
			"jnz	1f\n\t"
			"pushl	%4\n\t"		/* No */
			"movl	%5,%4\n"
		"3:	inb	%w2,%%al\n\t"
			LINK_SLOW_DOWN
			"andb	%6,%%al\n\t"
			"jnz	2f\n\t"
			"decl	%4\n\t"
			"jnz	3b\n\t"
			"popl	%4\n\t"
			"jmp	7f\n"		/* timeout */
		"2:	popl	%4\n"
		"1:	subb	%%ah,%b2\n\t"	/* set dx to IDR */
			"insb\n\t"		/* read one byte */
			LINK_SLOW_DOWN
			"addb	%%ah,%b2\n\t"	/* set it back to ISR */
			"decl	%4\n\t"		/* decrement bytes to go */
			"jnz	5b\n\t"
			"subl	%%esi,%0\n\t"
			"jz	8f\n\t"		/* all transfered */
			"pop	%%es\n\t"	/* after transfering LINK_MAX_BYTES */
			"pushl	%%edx\n\t"	/* it's time to call the */
			"call	link_call_sched\n\t"
			"popl	%%edx\n\t"
			"push	%%es\n\t"
			"movw	%%fs,%%ax\n\t"
			"movw	%%ax,%%es\n\t"
			"jmp	9b\n"
		"7:	subl	%4,%%esi\n\t"
			"subl	%%esi,%0\n"
		"8:	pop	%%es"

			:"=b" (count),
			 "=D" (temp)
			:"d" ((unsigned short) LINK_ISR(minor)),
			 "1" (temp),
			 "c" (count),
			 "i" (LINK_MAXTRY),
			 "i" (LINK_READBYTE),
			 "i" (LINK_MAX_BYTES)
			:"ax", "di", "cx", "dx", "si");
		DEB(link_total_bytes_read += temp - buf);

		if (!count) goto rquit;
#else
	while ( count )
	{
		end = (LINK_MAX_BYTES > count) ? count : LINK_MAX_BYTES;
		l_count = 0;
		do
		{
			if (!in(LINK_ISR(minor)) & LINK_READBYTE)
			{
				i = LINK_MAXTRY;
				while (!(in(LINK_ISR(minor)) & LINK_READBYTE) && i)
					i--;
				if (!i)
					break;
			}
			buffer[l_count++] = (char) in(LINK_IDR(minor));
			DEB(link_total_bytes_read++;)
		} while (--end);

		if (l_count)
		{
			copy_to_user(temp, buffer, l_count);
			count -= l_count;
			temp += l_count;
			link_call_sched_if();
		}
#endif /* LINK_USE_ASM */
		else while ( !( in(LINK_ISR(minor)) & LINK_READBYTE ) )
		{
			/* check for signals before going to sleep. */
			if (signal_pending(current)) {
				if ((temp - buf)==0) return -EINTR;
				goto rquit;
			}

			PRINTK("LINK(%d) sleeping at %d bytes for %d jiffies (reading).\n",
					minor, link_total_bytes_read, read_sleep );

			current->state = TASK_INTERRUPTIBLE;
			current->timeout = jiffies + read_sleep;
			link_call_sched();

			read_wait += read_sleep;
			read_inc = (read_inc + 1) % LINK_INC;
			read_sleep += (read_sleep < LINK_MAX_SLEEP && !read_inc);

			if ( read_wait > LINK_READ_TIMEOUT(minor)
				&& (LINK_F(minor) & LINK_READ_ABORT) )
			{
				PRINTK("LINK(%d): timeout during read.\n", minor);
				goto rquit;
			}
				
		}
	}
	PRINTKM("LINK(%d) reading successful.\n", minor);
rquit:
	*ppos += (temp - buf);
	return temp - buf;

} /* LINK_read() */



/****************************************************************************
 *
 * static int link_poll_write() - write to the link interface.
 *
 */
static ssize_t link_write(struct file *filp, const char *buf, size_t count,
		loff_t *ppos)
{
	const unsigned int minor = MINOR(filp->f_dentry->d_inode->i_rdev);
	unsigned int write_sleep = LINK_START_SLEEP;
	unsigned int write_wait  = 0;
	unsigned int write_inc = 0;
	DEB( unsigned int link_total_bytes_written = 0; )
#ifndef LINK_USE_ASM
	char	*temp = (char *)buf;
	int	l_count;
	int	end;
	char	buffer[LINK_MAX_BYTES];
	int	i;
#else
	register const char	*temp __asm__("si") = buf;
#endif

	PRINTKM("LINK(%d): writing %d bytes (polling mode).\n", minor, count);

	if (count < 0)
	{
		PRINTK("LINK invalid argument for writing: count = %d.\n", count);
		return -EINVAL;
	}

#ifdef LINK_USE_ASM
		__asm__("cld": :);

	while ( 1 )
	{
		/*
		 * this is very ugly!
		 */
		__asm__ __volatile__ (
			"movl	%4,%0\n\t"
			"push	%%ds\n\t"
			"push	%%fs\n\t"
			"movw	%%fs,%%ax\n\t"
			"movw	%%ds,%w4\n\t"
			"movw	%%ax,%%ds\n\t"
			"movw	%w4,%%fs\n"
		"9:	movl	%0,%4\n\t"
			"movb	$2,%%ah\n"
			"cmpl	%7,%4\n\t"
			"jc	6f\n\t"
			"movl	%7,%4\n"
		"6:	movl	%4,%%edi\n"
		"5:	inb	%w2, %%al\n\t"
			LINK_SLOW_DOWN
			"andb	%6,%%al\n\t"
			"jnz	1f\n\t"
			"pushl	%4\n\t"
			"movl	%5,%4\n"
		"3:	inb	%w2, %%al\n\t"
			LINK_SLOW_DOWN
			"andb	%6,%%al\n\t"
			"jnz	2f\n\t"
			"decl	%4\n\t"
			"jnz	3b\n\t"
			"popl	%4\n\t"
			"jmp	7f\n"
		"2:	popl	%4\n"
		"1:	subb	%%ah,%b2\n\t"
			"outsb\n\t"
			LINK_SLOW_DOWN
			"addb	%%ah,%b2\n\t"
			"decl	%4\n\t"
			"jnz	5b\n\t"
			"subl	%%edi,%0\n\t"
			"jz	8f\n\t"
			"pop	%%fs\n\t"	/* after transfering */
			"pop	%%ds\n\t"	/* LINK_MAX_BYTES it is time */
			"pushl	%%edx\n\t"	/* call the scheduler */
			"call	link_call_sched\n\t"
			"popl	%%edx\n\t"
			"push	%%ds\n\t"
			"push	%%fs\n\t"
			"movw	%%fs,%%ax\n\t"
			"movw	%%ds,%w4\n\t"
			"movw	%%ax,%%ds\n\t"
			"movw	%w4,%%fs\n\t"
			"jmp	9b\n"
		"7:	subl	%4,%%edi\n\t"
			"subl	%%edi,%0\n"
		"8:	pop	%%fs\n\t"
			"pop	%%ds"

			:"=b" (count),
			 "=S" (temp)
			:"d" ((unsigned short) LINK_OSR(minor)),
			 "1" (temp),
			 "c" (count),
			 "i" (LINK_MAXTRY),
			 "i" (LINK_WRITEBYTE),
			 "i" (LINK_MAX_BYTES)
			:"ax", "si", "dx", "cx", "di");
		DEB(link_total_bytes_written += temp - buf);

		if (!count)
			goto wquit;
#else
	while (count)
	{
		end = (LINK_MAX_BYTES > count) ? count : LINK_MAX_BYTES;
		copy_from_user(buffer, temp, end);
		l_count = 0;
		do
		{
			if (!in(LINK_OSR(minor)) & LINK_WRITEBYTE)
			{
				i = LINK_MAXTRY;
				while (!(in(LINK_OSR(minor)) & LINK_WRITEBYTE) && i)
					i--;
				if (!i)
					break;
			}
			out(buffer[l_count++], LINK_ODR(minor));
			DEB(link_total_bytes_written++;)
		} while (--end);

		if (l_count)
		{
			temp += l_count;
			count -= l_count;
			link_call_sched_if();
		}
#endif	/* LINK_USE_ASM */
		else while ( !( in(LINK_OSR(minor)) & LINK_WRITEBYTE ) )
		{
			/* check for signals before going to sleep. */
			if (signal_pending(current))
			{
				if (temp-buf) goto wquit; else return -EINTR;
			}

			PRINTK("LINK(%d) sleeping at %d bytes for %d jiffies (writing).\n",
					minor, link_total_bytes_written, write_sleep );

			current->state = TASK_INTERRUPTIBLE;
			current->timeout = jiffies + write_sleep;
			link_call_sched();

			write_wait += write_sleep;
			write_inc = (write_inc + 1) % LINK_INC;
			write_sleep += (write_sleep < LINK_MAX_SLEEP && !write_inc);

			if ( write_wait > LINK_WRITE_TIMEOUT(minor)
				&& (LINK_F(minor) & LINK_WRITE_ABORT) )
			{
				PRINTK("LINK(%d): timeout during write.\n", minor);
				goto wquit;
			}
				
		}
	}

	PRINTKM("LINK(%d): write successful.\n", minor);

wquit:
	link_boot_stat[minor] = LINK_STATUS_BOOTED;
	*ppos += (temp - buf);
	return temp-buf;
} /* link_poll_write() */

/****************************************************************************
 *
 * static int link_lseek()
 *
 */
static loff_t link_lseek(struct file * file, loff_t offset, int origin)
{
	switch (origin) {
		case 0:
			file->f_pos = offset;
			return file->f_pos;
		case 1:
			file->f_pos += offset;
			return file->f_pos;
		default:
			return -EINVAL;
	}
	if (file->f_pos < 0) return 0;
	return file->f_pos;
}


/****************************************************************************
 *
 * static int link_release()
 *
 * close the link device.
 *
 */
static int link_release(struct inode *inode, struct file *filp)
{
	const unsigned int minor = MINOR(inode->i_rdev);

	if (minor > 15) return 0;

	if (minor >= link_devices)
	{
		PRINTK("LINK(%d) not released, minor device number >= %d.\n",
			minor, link_devices);
		return -ENODEV;
	}

	LINK_F(minor) &= ~LINK_BUSY;

	MOD_DEC_USE_COUNT;

	PRINTK("LINK(%d) released.\n", minor);
	return 0;

} /* link_release() */


/****************************************************************************
 *
 * static int link_ioctl()
 *
 * This function performs the various ioctl() functions: resetting the
 * transputer, switching to analyse-mode, testing the status, changing
 * timeouts etc.
 *
 */
static int link_ioctl(struct inode *inode, struct file *file,
		    unsigned int cmd, unsigned long arg)
{
	const unsigned int minor = MINOR(inode->i_rdev);
	int result = arg;
	u_long k;

	PRINTK("LINK(%d) ioctl, cmd: 0x%08x, arg: 0x%08x.\n", minor, cmd, (int) arg);

	if (minor > 15) return -EINVAL;

	if ((minor >= link_devices) || !(LINK_F(minor) & LINK_BUSY) )
	{
		DEB(
			if (minor >= link_devices)
				printk("LINK ioctl exit, minor >= %d.\n", link_devices );
			else
				printk("LINK ioctl exit, device not opened.\n" );
		)
		return -ENODEV;
	}

	switch (cmd)
	{
		case LINKRESET:		/* reset transputer */
			link_reset(minor);
			break;
		case LINKWRITEABLE:	/* can we write a byte to the C012 ? */
	 		return ( ( in(LINK_OSR(minor)) & LINK_WRITEBYTE) != 0 ); 
		case LINKREADABLE:	/* can we read a byte from C012 ? */
	 		return ( ( in(LINK_ISR(minor)) & LINK_READBYTE) != 0 ); 
		case LINKANALYSE:	/* switch transputer to analyse mode */
			link_analyse(minor);
			break;
		case LINKERROR:		/* test error-flag */
			return ( in(LINK_BASE(minor) + LINK_ERROR_OFFSET) & LINK_TEST_ERROR) ? 0 : 1;
		case LINKREADTIMEOUT:	/* set timeout for reading */
			result = LINK_READ_TIMEOUT(minor);
			LINK_READ_TIMEOUT(minor) = arg;
			break;
		case LINKWRITETIMEOUT:	/* set timeout for writing */
			result = LINK_WRITE_TIMEOUT(minor);
			LINK_WRITE_TIMEOUT(minor) = arg;
			break;
		case LINKREADABORT:	/* abort after a timeout ? */
			if ( arg )
				LINK_F(minor) |= LINK_READ_ABORT;
			else
				LINK_F(minor) &= ~LINK_READ_ABORT;
			break;
		case LINKWRITEABORT:	/* abort after a timeout ? */
			if ( arg )
				LINK_F(minor) |= LINK_WRITE_ABORT;
			else
				LINK_F(minor) &= ~LINK_WRITE_ABORT;
			break;
		case LINKFASTREAD:
			if (in(LINK_ISR(minor)) & LINK_READBYTE)
			  result = (unsigned char) in(LINK_IDR(minor));
			else
			  result = -1;
			break;
		case LINKFASTWRITE:
			while (!in(LINK_OSR(minor)) & LINK_WRITEBYTE)
				link_call_sched_if();
			out((unsigned char) arg, LINK_ODR(minor));
			link_boot_stat[minor] = LINK_STATUS_BOOTED;
			break;
		case LINKPEEK:
			if (link_peek (minor, file->f_pos, &k))
				return -EINVAL;
			file->f_pos += 4;
			result = k;
			break;
		case LINKPOKE:
			if ((result=link_poke (minor, file->f_pos, arg)))
				return -EINVAL;
			file->f_pos += 4;
			break;
		case LINKGETSTAT:
			result = link_boot_stat[minor];
			break;
		default: result = -EINVAL;
	}

	PRINTK("LINK(%d) ioctl done.\n", minor);

	return result;

} /* link_ioctl() */


/****************************************************************************/
/* 2.1.60:
 * struct file_operations {
 *	loff_t (*llseek) (struct file *, loff_t, int);
 *	ssize_t (*read) (struct file *, char *, size_t, loff_t *);
 *	ssize_t (*write) (struct file *, const char *, size_t, loff_t *);
 *	int (*readdir) (struct file *, void *, filldir_t);
 *	unsigned int (*poll) (struct file *, poll_table *);
 *	int (*ioctl) (struct inode *, struct file *, unsigned int, unsigned long);
 *	int (*mmap) (struct file *, struct vm_area_struct *);
 *	int (*open) (struct inode *, struct file *);
 *	int (*release) (struct inode *, struct file *);
 *	int (*fsync) (struct file *, struct dentry *);
 *	int (*fasync) (struct file *, int);
 *	int (*check_media_change) (kdev_t dev);
 *	int (*revalidate) (kdev_t dev);
 *	int (*lock) (struct file *, int, struct file_lock *);
 * };
 */

static struct file_operations mem_fops = {
	NULL,		/* llseek */
	read_lmem,
	write_lmem,
	NULL,		/* readdir */
	NULL,		/* poll */
	NULL,		/* ioctl */
	NULL,		/* mmap */
	NULL,		/* open */
	link_release,
	NULL,		/* fsync */
	NULL,		/* fasync */
	NULL,		/* check_media_change */
	NULL,		/* revalidate */
	NULL		/* lock */
};

static struct file_operations link_fops = {
	link_lseek,
	link_read,
	link_write,
	NULL,		/* readdir */
	NULL,		/* poll */
	link_ioctl,
	NULL,		/* mmap */
	NULL,		/* open */
	link_release,
	NULL,		/* fsync */
	NULL,		/* fasync */
	NULL,		/* check_media_change */
	NULL,		/* revalidate */
	NULL		/* lock */
};


/****************************************************************************
 *
 * static int link_open()
 *
 * open the link-device.
 *
 */
static int link_open(struct inode *inode, struct file *filp)
{
	const unsigned int minor = MINOR(inode->i_rdev);

	if (minor > 15) {
		filp->f_op = &mem_fops;
		return 0;
	}

	if (minor >= link_devices)
	{
		PRINTK("LINK(%d) not opened, minor device number >= %d.\n",
			minor, link_devices);
		return -ENODEV;
	}

	if (LINK_F(minor) & LINK_BUSY)
	{
		PRINTK("LINK(%d) not opened, device busy.\n", minor);
		return -EBUSY;
	}

	LINK_F(minor) |= LINK_BUSY;
	filp->f_op = &link_fops;

	MOD_INC_USE_COUNT;

	PRINTK( "LINK(%d) opened.\n", minor);
	return 0;

} /* link_open() */


/****************************************************************************
 *
 * link_fops
 *
 * File operations
 */
static struct file_operations dtc2l_fops = {
	NULL,		/* llseek */
	NULL,		/* read */
	NULL,		/* write */
	NULL,		/* readdir */
	NULL,		/* poll */
	NULL,		/* ioctl */
	NULL,		/* mmap */
	link_open,
	NULL,		/* release */
	NULL,		/* fsync */
	NULL,		/* fasync */
	NULL,		/* check_media_change */
	NULL,		/* revalidate */
	NULL		/* lock */
};


/****************************************************************************
 *
 * long link_init()
 *
 * This function initializes the driver. It tries to detect the hardware
 * and sets up all relevant data structures.
 *
 */
long link_init(long kmem_start)
{
	unsigned int test = 0, i;
	unsigned char link_dev = 0;

	if ( register_chrdev( LINK_MAJOR, LINK_NAME, &dtc2l_fops ) )
	{
		printk("link_init: unable to get major %d for link interface.\n",
			LINK_MAJOR );
		return kmem_start;
	}

	/*
	   After a reset it should be possible to write a byte to
	   the LINK. So let's do a reset and then test the output status
	   register
	*/
	for (; link_base_addresses[test] && link_dev < LINK_NO; test++)
	{
		udelay(LINK_DELAY * 10);
		LINK_BASE(link_dev) = link_base_addresses[test];
		LINK_ODR(link_dev) = LINK_BASE(link_dev) + LINK_ODR_OFFSET;
		LINK_ISR(link_dev) = LINK_BASE(link_dev) + LINK_ISR_OFFSET;
		LINK_OSR(link_dev) = LINK_BASE(link_dev) + LINK_OSR_OFFSET;

		link_reset(link_dev);

		udelay(LINK_DELAY * 10);

		for (i = 0; i < LINK_MAXTRY; i++) if ( in(LINK_OSR(link_dev)) == LINK_WRITEBYTE)
		{
			out(B008_INT(link_dev), 0);
			udelay(LINK_DELAY * 10);
			LINK_BOARDTYPE(link_dev) =
				(in(B008_INT(link_dev)) & 0x0f) == 0 ?
					LINK_B008 : LINK_B004;
/*restart_init:*/
			/* LINK_BOARDTYPE(link_dev) = LINK_B008; */
			printk("link%d at 0x0%x (polling) is a B00%c\n",
				link_dev,LINK_IDR(link_dev),
				LINK_BOARDTYPE(link_dev));

			request_region(LINK_IDR(link_dev), 
				LINK_BOARDTYPE(link_dev) == LINK_B004 ? B004_IO_SIZE : B008_IO_SIZE,
				LINK_NAME);
			link_dev++;
			break;
		}

		}

	if (!link_dev)
		printk("link: no interfaces found.\n");

	link_devices = link_dev;

	return kmem_start;

} /* link_init() */

#ifdef MODULE
/****************************************************************************
 *
 * init_module()
 *
 * Initialize the driver when loaded as a module
 */
int init_module(void)
{
	return link_init(0L);
} /* init_module() */

/****************************************************************************
 *
 * cleanup_module()
 *
 * Prepare to remove the driver
 */
void cleanup_module(void)
{
	int i = link_devices;
	if (MOD_IN_USE)
		printk("link: device busy, remove delayed.\n");
	else
	{
		unregister_chrdev( LINK_MAJOR, LINK_NAME );
		for (; (--i) >= 0; )
		{
			release_region(LINK_IDR(i),
				LINK_BOARDTYPE(i) ==
				LINK_B004 ? B004_IO_SIZE : B008_IO_SIZE);
		}
	}
} /* cleanup_module() */

#endif /* MODULE */
