/************************************************************************
*									*
*	Device Driver for INMOS-B004/B008 compatible link interfaces	*
*									*
*	Copyright (C) 1993,94,95 by Christoph Niemann			*
*				    niemann@swt.ruhr-uni-bochum.de	*
*	(loosely) based on						*
*		usr/include/linux/lp.h c.1991-1992 James Wiegand	*
*		many modifications copyright (C) 1992 Michael K. Johnson*
*									*
*	Version 0.9, for Linux 1.1.83 and above, February 1995.		*
*									*
*	The driver may be copied under the same conditions as the	*
*	Linux operating system						*
*									*
*************************************************************************/

#ifndef _LINUX_LINK_H
#define _LINUX_LINK_H

#include <linux/timer.h>

/*
 * Per POSIX guidelines, this module reserves the LINK and link prefixes
 */

#define LINK_MAJOR	31		/* Major device number */
#define LINK_NAME	"link"		/* Name of the device to register */
#define LINK_NO		2		/* Number of supported boards */

/*
 * Transfer mode to use with a board that supports DMA
 */
#define LINK_USE_PIO	0

/*
 * Delay for resetting the transputer system in usecs
 */
#define LINK_DELAY	100L

/*
 * These are the link_table[].flags flags...
 */
#define LINK_EXIST 		0x0001	/* Is a B004-Board with at least one 
					   Transputer present ? */
#define LINK_BUSY  		0x0002	/* Is the B004-board in use ? */
#define LINK_READ_ABORT		0x0004  /* abort reading after timeout ? */
#define LINK_WRITE_ABORT	0x0008  /* abort writing after timeout ? */
#define LINK_READ_DMA		0x0010	/* DMA read is active */
#define LINK_WRITE_DMA		0x0020	/* DMA write is active */
#define LINK_IN_INTERRUPT	0x0040	/* set inside IRQ handler */

/*
 * IOCTL numbers
 */
#define LINKTIME		0x0001		/* no longer available */
#define LINKRESET		0x0002		/* reset transputer */
#define LINKWRITEABLE		0x0004		/* check if we can send a byte */
#define LINKREADABLE		0x0008		/* check if we can receive a byte */
#define LINKANALYSE		0x0010		/* go to analyse mode */
#define LINKERROR		0x0020
#define LINKREADTIMEOUT		0x0040		/* set timeout for reading */
#define LINKWRITETIMEOUT	0x0080		/* set timeout for writing */
#define LINKREADABORT		0x0100
#define LINKWRITEABORT		0x0200
#define LINKFASTREAD		0x0400
#define LINKFASTWRITE		0x0800
#define LINKPEEK		0x1000
#define LINKPOKE		0x2000
#define LINKGETSTAT		0x4000

/*
 * timeout for printk'ing a timeout, in jiffies (100ths of a second).
 * This is also used for re-checking error conditions if LINK_READ_ABORT or 
 * LINK_WRITE_ABORT is not set. This is the default behavior for reading.
 * Writing has the timeout enabled per default.
 */
#define LINK_INIT_WRITE_TIMEOUT	1000
#define LINK_INIT_READ_TIMEOUT	100

/*
 * If the link interface is not ready for reading or writing the driver starts
 * to poll the interface.
 * At the begining, the driver sleeps for LINK_START_SLEEP jiffies. If the
 * link interface is still unable to send or receive a byte, the driver
 * sleeps again for the same duration.
 * After LINK_INC times sleeping LINK_START_SLEEP jiffies, the driver adds one
 * jiffy to its sleeping time. After LINK_INC times sleeping the new amount
 * of time, it is incremented again and so on.
 * The maximum time to sleep without rechecking the link interface is specified
 * by LINK_MAX_SLEEP.
 */

#define LINK_MAX_SLEEP	20
#define LINK_START_SLEEP 1
#define LINK_INC	16

/*
 * The addresses of the C012 hardware registers relative to the
 * base address.
 */

#define LINK_IDR_OFFSET		0		/* Input Data Register */
#define LINK_ODR_OFFSET		1		/* Output Data Register */
#define LINK_ISR_OFFSET		2		/* Input Status Register */
#define LINK_OSR_OFFSET		3		/* Output Status Register */
#define LINK_RESET_OFFSET	16		/* Reset/Error Register */
#define LINK_ERROR_OFFSET	LINK_RESET_OFFSET
#define LINK_ANALYSE_OFFSET	17		/* Analyse Register */
#define B008_DMA_OFFSET		18		/* B008: DMA request register */
#define B008_INT_OFFSET		19		/* B008: Interrupt control reg */
#define B004_IO_SIZE		18
#define B008_IO_SIZE		20

/*
 * Minimum number of bytes to be transferred using DMA
 */
#define B008_DMA_MIN		512

#ifndef B008_DMA_CHANNEL
#define B008_DMA_CHANNEL	3
#endif
#ifndef B008_IRQ_LINE
#define B008_IRQ_LINE		3
#endif

struct link_struct {
	int flags;			/* various flags */
	int idr;			/* address of the input data register */
	int odr;			/* address if the output data register */
	int isr;			/* address of the input status register */
	int osr;			/* address of the output status register */
	unsigned int read_timeout;	/* timeout for reading from the link */
	unsigned int write_timeout;	/* timeout for writing to the link */
	int boardtype;			/* what kind of board is installed */
#ifdef B008_SUPPORT
	unsigned char dma;		/* dma channel to use */
	unsigned char irq;		/* used irq */
#endif
};

/*
 * Id's for the supported boards
 */
#define LINK_B004 '4'
#define LINK_B008 '8'

/*
 * Defines for easier access to the link_table.
 */
#define LINK_F(minor)			link_table[(int) minor].flags
#define LINK_READ_TIMEOUT(minor)	link_table[(int) minor].read_timeout
#define LINK_WRITE_TIMEOUT(minor)	link_table[(int) minor].write_timeout
#define LINK_BASE(minor)		LINK_IDR(minor)
#define LINK_IDR(minor)			link_table[(int) minor].idr
#define LINK_ODR(minor)			link_table[(int) minor].odr
#define LINK_ISR(minor)			link_table[(int) minor].isr
#define LINK_OSR(minor)			link_table[(int) minor].osr
#define LINK_WAIT(minor)		link_table[(int) minor].wait
#define LINK_BOARDTYPE(minor)		link_table[(int) minor].boardtype
#define LINK_RESET(minor)		((LINK_BASE(minor) + LINK_RESET_OFFSET) & 0x3ff)
#define LINK_ANALYSE(minor)		((LINK_BASE(minor) + LINK_ANALYSE_OFFSET) & 0x3ff)
/*
 * Address of the interrupt register
 */
#define B008_INT(minor)			(LINK_BASE(minor) + B008_INT_OFFSET)
#ifdef B008_SUPPORT
/*
 * Address of the DMA control register
 */
#define B008_DMA(minor)			(LINK_BASE(minor) + B008_DMA_OFFSET)
#define LINK_DMA(minor)			link_table[(int) minor].dma
#define LINK_IRQ(minor)			link_table[(int) minor].irq
#endif

/*
 * Number of tries to access isr or osr before reading or writing fails
 */
#define LINK_MAXTRY 			300

/*
 * Maximum number of bytes to transfer without calling the scheduler
 */
#ifdef LINK_USE_ASM
#define LINK_MAX_BYTES			32768
#else
#define LINK_MAX_BYTES			256
#endif

/*
 * bit defines for C012 status ports at base + 2/3
 * accessed with LINK_IS, LINK_OS, which gets the byte...
 */
#define LINK_READBYTE	1
#define LINK_WRITEBYTE	1

/*
 * Values to write into the DMA control register of the B008 to start
 * a DMA write or read
 */
#define LINK_START_DMA_READ	1
#define LINK_START_DMA_WRITE	0

/*
 * bit defines for C012 reset/error port at base + 16
 */
#define LINK_ASSERT_RESET	0x01	/* resetting the transputer */
#define LINK_DEASSERT_RESET	0x00
#define LINK_TEST_ERROR		0x01	/* for testing the transputer's error flag */

/*
 * bit defines for C012 analyse port at base + 17
 */
#define LINK_ASSERT_ANALYSE	0x01	/* switch transputer to analyse-mode */
#define LINK_DEASSERT_ANALYSE	0x00

/*
 * The interrupt control register on a B008 compatible board is used to
 * determine which of the following events cause an interrupt
 */
#define LINK_INT_DMA		0x01	/* end of DMA transfer */
#define LINK_INT_ERR		0x02	/* transputer's error flags is set */
#define LINK_INT_RECIVE		0x04	/* B008 is ready to receive a byte */
#define LINK_INT_SEND		0x08	/* B008 is ready to send a byte */

/*
 */

#define LINK_STATUS_RESET	0x01
#define LINK_STATUS_ANALYSE	0x02
#define LINK_STATUS_BOOTED	0x00

#endif
